@@ -38,8 +38,7 @@
      data may linger in the database. We periodically sweep the
      database, compacting it and performing regular cleanup so we can
      maintain acceptable performance."
-  (:require [clj-time.core :refer [ago]]
-            [clojure.java.io :as io]
+  (:require [clojure.java.io :as io]
             [clojure.string :as str]
             [clojure.tools.logging :as log]
             [compojure.core :as compojure]
@@ -66,7 +65,7 @@
             [puppetlabs.puppetdb.scf.storage :as scf-store]
             [puppetlabs.puppetdb.scf.storage-utils :as sutils]
             [puppetlabs.puppetdb.schema :as pls :refer [defn-validated]]
-            [puppetlabs.puppetdb.time :refer [to-seconds to-millis parse-period
+            [puppetlabs.puppetdb.time :refer [ago to-seconds to-millis parse-period
                                               format-period period?]]
             [puppetlabs.puppetdb.utils :as utils]
             [puppetlabs.trapperkeeper.core :refer [defservice] :as tk]
@@ -333,31 +332,53 @@
         migrated?))))
 
 (defn init-with-db
-  "All initialization operations needing a database connection should
-  happen here. This function creates a connection pool using
-  `write-db-config` that will hang until it is able to make a
-  connection to the database. This covers the case of the database not
-  being fully started when PuppetDB starts. This connection pool will
-  be opened and closed within the body of this function.  Returns true
-  iff any migrations were run.  Throws
-  {:type ::unsupported-database :current version :oldest version} if the
-  current database is not supported."
+  "Performs all initialization operations requiring a database
+  connection using a transient connection pool derived from the
+  `write-db-config`.  Blocks until the initialization is complete, and
+  then returns an unspecified value.  Throws exceptions on errors.
+  Throws {:type ::unsupported-database :current version :oldest
+  version} if the current database is not supported."
   [write-db-config config]
-  (loop [db-spec (assoc write-db-config
-                        ;; Block waiting to grab a connection
-                        :connection-timeout 15000
-                        :pool-name "PDBMigrationsPool")]
-    (if-let [result
-             (try
-               (with-open [init-db-pool (jdbc/make-connection-pool db-spec
-                                                                   database-metrics-registry)]
-                 (let [db-pool-map {:datasource init-db-pool}]
-                   (initialize-schema db-pool-map config)
-                   ::success))
-               (catch java.sql.SQLTransientConnectionException e
-                 (log/error e (trs "Error while attempting to create connection pool"))))]
-      result
-      (recur db-spec))))
+  ;; A C-c (SIGINT) will close the pool via the shutdown hook, which
+  ;; will then cause the pool connection to throw a (generic)
+  ;; SQLException.
+  (with-open [db-pool (-> (assoc write-db-config
+                                 :pool-name "PDBMigrationsPool"
+                                 :connection-timeout 3000)
+                          (jdbc/make-connection-pool database-metrics-registry))]
+    (let [runtime (Runtime/getRuntime)
+          on-shutdown (doto (Thread.
+                             #(try
+                                (.close db-pool)
+                                (catch Exception ex
+                                  ;; Nothing to do but log it...
+                                  (log/error ex (trs "Unable to close migration pool")))))
+                        (.setName "PuppetDB migration pool closer"))]
+      (.addShutdownHook runtime on-shutdown)
+      (try
+        (loop [i 0
+               last-ex nil]
+          (let [result (try
+                         (initialize-schema {:datasource db-pool} config)
+                         true
+                         (catch java.sql.SQLTransientConnectionException ex
+                           ;; When coupled with the 3000ms timout, this
+                           ;; is intended to log a duplicate message no
+                           ;; faster than once per minute.
+                           (when (or (not= (str ex) (str last-ex))
+                                     (zero? (mod i 20)))
+                             (log/error (str (trs "Will retry database connection after temporary failure: ")
+                                             ex)))
+                           ex))]
+            (if (true? result)
+              result
+              (recur (inc i) result))))
+        (finally
+          (try
+            (.removeShutdownHook runtime on-shutdown)
+            (catch IllegalStateException ex
+              ;; Ignore, because we're already shutting down.
+              nil)))))))
 
 (defn db-config->clean-request
   [config]
@@ -400,6 +421,13 @@
       (finally
         (swap! stop-status disj :collecting-garbage)))))
 
+(defn maybe-send-cmd-event!
+  "Put a map with identifying information about an enqueued command on the
+   cmd-event-chan. Valid :kind values are: ::command/ingested ::command/processed"
+  [emit-cmd-events? cmd-event-ch cmdref kind]
+  (when emit-cmd-events?
+    (async/>!! cmd-event-ch (queue/make-cmd-event cmdref kind))))
+
 (defn start-puppetdb
   "Throws {:type ::unsupported-database :current version :oldest version} if
   the current database is not supported."
@@ -409,10 +437,12 @@
    :post [(map? %)]}
   (let [{:keys [developer jetty
                 database read-database
-                puppetdb command-processing]} config
+                puppetdb command-processing
+                emit-cmd-events?]} config
         {:keys [pretty-print max-enqueued]} developer
         {:keys [gc-interval]} database
         {:keys [disable-update-checking]} puppetdb
+        {:keys [cmd-event-mult cmd-event-ch]} context
 
         write-db (jdbc/pooled-datasource (assoc database :pool-name "PDBWritePool")
                                          database-metrics-registry)
@@ -433,7 +463,11 @@
                          (queue/sorted-command-buffer
                           max-enqueued
                           #(cmd/update-counter! :invalidated %1 %2 inc!)))
-          [q load-messages] (queue/create-or-open-stockpile (conf/stockpile-dir config))
+          emit-cmd-events? (or (conf/pe? config) emit-cmd-events?)
+          maybe-send-cmd-event! (partial maybe-send-cmd-event! emit-cmd-events? cmd-event-ch)
+          [q load-messages] (queue/create-or-open-stockpile (conf/stockpile-dir config)
+                                                            maybe-send-cmd-event!
+                                                            cmd-event-ch)
           globals {:scf-read-db read-db
                    :scf-write-db write-db
                    :pretty-print pretty-print
@@ -441,12 +475,18 @@
                    :dlo (dlo/initialize (get-path stockdir "discard")
                                         (get-in metrics-registries
                                                 [:dlo :registry]))
-                   :command-chan command-chan}
+                   :command-chan command-chan
+                   :cmd-event-mult cmd-event-mult
+                   :maybe-send-cmd-event! maybe-send-cmd-event!}
           clean-lock (ReentrantLock.)
           command-loader (when load-messages
                            (future
                              (load-messages command-chan cmd/inc-cmd-depth)))]
 
+      ;; send the queue-loaded cmd-event if the command-loader didn't
+      (when-not command-loader
+        (async/>!! cmd-event-ch {:kind ::queue/queue-loaded}))
+
       ;; Pretty much this helper just knows our job-pool and gc-interval
       (let [job-pool (mk-pool)
             gc-interval-millis (to-millis gc-interval)]
@@ -488,7 +528,16 @@
     Returns false if some kind of maintenance was already in progress,
     true otherwise.  Although the latter does not imply that all of
     the operations were successful; consult the logs for more
-    information."))
+    information.")
+  (cmd-event-mult [this]
+    "Returns a core.async mult to which
+     {:kind <::ingested|::processed|::queue/queue-loaded>} maps are sent
+     when emit-cmd-events? is set to true. In the case of
+     {:kind <::ingested|::processed>} the map will contain :entity, :certname,
+     and :producer-ts entries for each command PDB has either ingested or
+     finished processing. The {:kind ::queue/queue-loaded} map is used to
+     indicate that the queue/message-loader has finished loading any existing
+     commands from the stockpile queue when PDB starts up."))
 
 (defservice puppetdb-service
   "Defines a trapperkeeper service for PuppetDB; this service is responsible
@@ -502,12 +551,16 @@
 
         (doseq [{:keys [reporter]} (vals metrics/metrics-registries)]
           (jmx-reporter/start reporter))
-        (assoc context
-               :url-prefix (atom nil)
-               ;; This coordination is needed until we no longer
-               ;; support jdk < 10.
-               ;; https://bugs.openjdk.java.net/browse/JDK-8176254
-               :stop-status (atom #{})))
+        (let [cmd-event-ch (async/chan 1000)
+              cmd-event-mult (async/mult cmd-event-ch)]
+          (assoc context
+                 :cmd-event-ch cmd-event-ch
+                 :cmd-event-mult cmd-event-mult
+                 :url-prefix (atom nil)
+                 ;; This coordination is needed until we no longer
+                 ;; support jdk < 10.
+                 ;; https://bugs.openjdk.java.net/browse/JDK-8176254
+                 :stop-status (atom #{}))))
   (start [this context]
          (try+
           (start-puppetdb context (get-config) this get-registered-endpoints)
@@ -524,7 +577,13 @@
   (stop [this context]
         (doseq [{:keys [reporter]} (vals metrics/metrics-registries)]
           (jmx-reporter/stop reporter))
+        (let [{:keys [cmd-event-mult cmd-event-ch]} context]
+          (when cmd-event-mult
+            (async/untap-all cmd-event-mult))
+          (when cmd-event-ch
+            (async/close! cmd-event-ch)))
         (stop-puppetdb context))
+
   (set-url-prefix [this url-prefix]
                   (let [old-url-prefix (:url-prefix (service-context this))]
                     (when-not (compare-and-set! old-url-prefix nil url-prefix)
@@ -545,7 +604,9 @@
                                      row-callback-fn)))
 
   (clean [this] (clean this #{}))
-  (clean [this what] (clean-puppetdb (service-context this) (get-config) what)))
+  (clean [this what] (clean-puppetdb (service-context this) (get-config) what))
+
+  (cmd-event-mult [this] (-> this service-context :cmd-event-mult)))
 
 (defn provide-services
   "Starts PuppetDB as a service via Trapperkeeper.  Augments TK's normal
