@@ -62,6 +62,10 @@
             [puppetlabs.puppetdb.cheshire :as json]
             [puppetlabs.puppetdb.jdbc :as jdbc]
             [puppetlabs.puppetdb.schema :refer [defn-validated]]
+            [puppetlabs.puppetdb.time :as fmt-time]
+            [puppetlabs.puppetdb.time :as tcoerce]
+            [puppetlabs.puppetdb.time :as time
+             :refer [now in-millis interval to-timestamp]]
             [puppetlabs.puppetdb.utils :as utils]
             [slingshot.slingshot :refer [try+ throw+]]
             [puppetlabs.puppetdb.command.constants
@@ -70,9 +74,7 @@
              :refer [defservice service-context]]
             [schema.core :as s]
             [puppetlabs.puppetdb.config :as conf]
-            [puppetlabs.puppetdb.time :refer [to-timestamp]]
-            [clj-time.core :as time :refer [now interval in-millis]]
-            [clj-time.format :as fmt-time]
+            [puppetlabs.puppetdb.time :as time]
             [clojure.set :as set]
             [clojure.core.async :as async]
             [metrics.timers :refer [timer time!]]
@@ -81,7 +83,6 @@
             [metrics.histograms :refer [histogram update!]]
             [puppetlabs.puppetdb.metrics.core :as metrics]
             [puppetlabs.puppetdb.queue :as queue]
-            [clj-time.coerce :as tcoerce]
             [puppetlabs.puppetdb.amq-migration :as mig]
             [puppetlabs.puppetdb.command.dlo :as dlo]
             [overtone.at-at :as at-at :refer [mk-pool stop-and-reset-pool!]]
@@ -248,13 +249,15 @@
   [q
    command-chan
    ^Semaphore write-semaphore
-   {:keys [command certname command-stream compression] :as command-req} :- queue/command-req-schema]
+   {:keys [command certname command-stream compression] :as command-req} :- queue/command-req-schema
+   maybe-send-cmd-event!]
   (try
     (.acquire write-semaphore)
     (time! (get @metrics :message-persistence-time)
-           (let [cmd (queue/store-command q command-req)
-                 {:keys [id received]} cmd]
-             (async/>!! command-chan cmd)
+           (let [cmdref (queue/store-command q command-req)
+                 {:keys [id received]} cmdref]
+             (async/>!! command-chan cmdref)
+             (maybe-send-cmd-event! cmdref ::ingested)
              (log/debug (trs "[{0}-{1}] ''{2}'' command enqueued for {3}"
                              id
                              (tcoerce/to-long received)
@@ -287,8 +290,8 @@
 ;; Catalog replacement
 
 (defn replace-catalog*
-  [{:keys [certname version id received payload]} start-time db]
-  (let [{producer-timestamp :producer_timestamp :as catalog} payload]
+  [{:keys [version id received payload]} start-time db]
+  (let [{producer-timestamp :producer_timestamp certname :certname :as catalog} payload]
     (jdbc/with-transacted-connection' db :repeatable-read
       (scf-storage/maybe-activate-node! certname producer-timestamp)
       (scf-storage/replace-catalog! catalog received))
@@ -488,10 +491,11 @@
 
 (defn discard-message
   "Discards the given `cmdref` caused by `ex`"
-  [cmdref ex q dlo]
+  [cmdref ex q dlo maybe-send-cmd-event!]
   (-> cmdref
       (queue/cons-attempt ex)
       (dlo/discard-cmdref q dlo))
+  (maybe-send-cmd-event! cmdref ::processed)
   (let [{:keys [command version]} cmdref]
     (mark-both-metrics! command version :discarded)
     (update-counter! :depth command version dec!)))
@@ -501,11 +505,12 @@
   processing a non-delete cmdref except different metrics need to be
   updated to indicate the difference in command"
   [{:keys [command version certname id received] :as cmdref}
-   q scf-write-db response-chan stats blacklist-config]
+   q scf-write-db response-chan stats blacklist-config maybe-send-cmd-event!]
   (process-command-and-respond! cmdref scf-write-db response-chan stats blacklist-config)
   (log-command-processed-messsage id received (now) (command-keys command)
                                   certname {:obsolete-cmd? true})
   (queue/ack-command q {:entry (queue/cmdref->entry cmdref)})
+  (maybe-send-cmd-event! cmdref ::processed)
   (update-counter! :depth command version dec!)
   (update-counter! :invalidated command version dec!))
 
@@ -513,7 +518,7 @@
   "Parses, processes and acknowledges a successful command ref and
   updates the relevant metrics. Any exceptions that arise are
   unhandled and expected to be caught by the caller."
-  [cmdref q scf-write-db response-chan stats blacklist-config]
+  [cmdref q scf-write-db response-chan stats blacklist-config maybe-send-cmd-event!]
   (let [{:keys [command version] :as cmd} (queue/cmdref->cmd q cmdref)
         retries (count (:attempts cmdref))]
     (if-not cmd
@@ -522,7 +527,8 @@
         (create-metrics-for-command! command version)
         (mark-both-metrics! command version :seen)
         (update-counter! :depth command version dec!)
-        (mark! (global-metric :fatal)))
+        (mark! (global-metric :fatal))
+        (maybe-send-cmd-event! cmdref ::processed))
       (do
         (create-metrics-for-command! command version)
 
@@ -538,6 +544,7 @@
          (mark-both-metrics! command version :processed))
 
         (queue/ack-command q cmd)
+        (maybe-send-cmd-event! cmdref ::processed)
         (update-counter! :depth command version dec!)))))
 
 (def command-delay-ms (* 1000 60 60))
@@ -581,30 +588,32 @@
   that fail via (delay-message msg), and discarding messages that have
   fatal errors or have exceeded their maximum allowed attempts."
   [q command-chan dlo delay-pool scf-write-db response-chan stats
-   blacklist-config stop-status]
+   blacklist-config stop-status maybe-send-cmd-event!]
   (fn [{:keys [certname command version received delete? id] :as cmdref}]
     (try+
 
      (when received
        (let [q-time (-> (fmt-time/parse iso-formatter received)
-                        (time/interval (time/now))
+                        (time/interval (now))
                         time/in-seconds)]
          (create-metrics-for-command! command version)
          (update! (global-metric :queue-time) q-time)
          (update! (cmd-metric command version :queue-time) q-time)))
 
      (if delete?
-       (process-delete-cmdref cmdref q scf-write-db response-chan stats blacklist-config)
+       (process-delete-cmdref cmdref q scf-write-db response-chan
+                              stats blacklist-config maybe-send-cmd-event!)
        (let [retries (count (:attempts cmdref))]
          (try+
-          (process-cmdref cmdref q scf-write-db response-chan stats blacklist-config)
+          (process-cmdref cmdref q scf-write-db response-chan stats
+                          blacklist-config maybe-send-cmd-event!)
           (catch fatal? obj
             (mark! (global-metric :fatal))
             (let [ex (:cause obj)]
               (log/error
                (:wrapper &throw-context)
                (trs "[{0}] [{1}] Fatal error on attempt {2} for {3}" id command retries certname))
-              (discard-message cmdref ex q dlo)))
+              (discard-message cmdref ex q dlo maybe-send-cmd-event!)))
           (catch Exception _
             (let [ex (:throwable &throw-context)]
               (mark-both-metrics! command version :retried)
@@ -619,13 +628,13 @@
                   (log/error
                    ex
                    (trs "[{0}] [{1}] Exceeded max attempts ({2}) for {3}" id command retries certname))
-                  (discard-message cmdref ex q dlo))))))))
+                  (discard-message cmdref ex q dlo maybe-send-cmd-event!))))))))
 
      (catch [:kind ::queue/parse-error] _
        (mark! (global-metric :fatal))
        (log/error (:wrapper &throw-context)
                   (trs "Fatal error parsing command: {0}" (:id cmdref)))
-       (discard-message cmdref (:throwable &throw-context) q dlo)))))
+       (discard-message cmdref (:throwable &throw-context) q dlo maybe-send-cmd-event!)))))
 
 (def stop-commands-wait-ms (constantly 5000))
 
@@ -677,7 +686,7 @@
          ;; which means we've done everything that needs to be done.
          (request-shutdown)
          context)
-       (let [{:keys [command-chan scf-write-db q]} globals
+       (let [{:keys [command-chan scf-write-db q maybe-send-cmd-event!]} globals
              {:keys [response-chan response-pub]} context
              ;; From mq_listener
              command-threadpool (create-command-handler-threadpool (conf/mq-thread-count config))
@@ -693,7 +702,8 @@
                                              :database
                                              (select-keys [:facts-blacklist
                                                            :facts-blacklist-type]))
-                                         (:stop-status context))]
+                                         (:stop-status context)
+                                         maybe-send-cmd-event!)]
 
          (doto (Thread. (fn []
                           (try+ (gtp/dochan command-threadpool handle-cmd command-chan)
@@ -735,13 +745,14 @@
                    (enqueue-command this command version certname producer-ts command-stream compression identity))
 
   (enqueue-command [this command version certname producer-ts command-stream compression command-callback]
-    (let [config (get-config)
-          q (:q (shared-globals))
-          command-chan (:command-chan (shared-globals))
-          write-semaphore (:write-semaphore (service-context this))
-          command (if (string? command) command (command-names command))
-          command-req (queue/create-command-req command version certname producer-ts compression command-callback command-stream)
-          result (do-enqueue-command q command-chan write-semaphore command-req)]
+   (let [config (get-config)
+         globals (shared-globals)
+         q (:q globals)
+         command-chan (:command-chan globals)
+         write-semaphore (:write-semaphore (service-context this))
+         command (if (string? command) command (command-names command))
+         command-req (queue/create-command-req command version certname producer-ts compression command-callback command-stream)
+         result (do-enqueue-command q command-chan write-semaphore command-req (:maybe-send-cmd-event! globals))]
       ;; Obviously assumes that if do-* doesn't throw, msg is in
       (inc-cmd-depth command version)
       (swap! (:stats (service-context this)) update :received-commands inc)
