@@ -7,7 +7,6 @@
            [org.apache.commons.compress.compressors.gzip GzipCompressorInputStream])
   (:require [clojure.string :as str :refer [re-quote-replacement]]
             [puppetlabs.stockpile.queue :as stock]
-            [clj-time.coerce :as tcoerce]
             [clojure.tools.logging :as log]
             [puppetlabs.i18n.core :refer [trs tru]]
             [puppetlabs.puppetdb.cheshire :as json]
@@ -15,7 +14,6 @@
             [puppetlabs.puppetdb.constants :as constants]
             [metrics.timers :refer [timer time!]]
             [metrics.counters :refer [inc!]]
-            [clj-time.core :as time]
             [puppetlabs.kitchensink.core :as kitchensink]
             [slingshot.slingshot :refer [throw+]]
             [clojure.core.async :as async]
@@ -27,8 +25,10 @@
                                                utf8-truncate]]
             [slingshot.slingshot :refer [try+]]
             [schema.core :as s]
+            [puppetlabs.puppetdb.time :as tcoerce]
+            [puppetlabs.puppetdb.time :as time]
             [puppetlabs.puppetdb.schema :as pls]
-            [puppetlabs.puppetdb.time :as pdbtime]))
+            [puppetlabs.puppetdb.time :refer [now parse-wire-datetime]]))
 
 (def metadata-command->puppetdb-command
   ;; note that if there are multiple metadata names for the same command then
@@ -233,8 +233,7 @@
   {:command command
    :version version
    :certname certname
-   :producer-ts (when producer-ts
-                  (pdbtime/from-string producer-ts))
+   :producer-ts (some-> producer-ts parse-wire-datetime)
    :compression compression
    :callback callback
    :command-stream command-stream})
@@ -341,7 +340,7 @@
 (s/defn store-command
   [q
    command-req :- command-req-schema]
-  (let [current-time (time/now)
+  (let [current-time (now)
         entry (store-in-stockpile q
                                   current-time
                                   command-req)]
@@ -427,6 +426,17 @@
    ;; to get rid of it somehow when we refactor mq-listener and command.clj.
    (SortedCommandBuffer. (TreeMap.) (HashMap.) n delete-update-fn)))
 
+(defn make-cmd-event
+  "Given a cmdref and kind return a cmd-event-map which is suitable to be put
+   on the cmd-event-chan.
+   Valid :kind values are: ::command/ingested and ::command/processed."
+  [cmdref kind]
+  (let [{:keys [command certname producer-ts]} cmdref]
+    {:kind kind
+     :command (command-constants/command-keys command)
+     :certname certname
+     :producer-ts producer-ts}))
+
 (defn message-loader
   "Returns a function that will enqueue existing stockpile messages to
   `command-chan`. Messages with ids less than `message-id-ceiling`
@@ -434,29 +444,32 @@
   when new commands are enqueued before all existing commands have
   been enqueued. Note that there is no guarantee on the enqueuing
   order of commands read from stockpile's reduce function"
-  [q message-id-ceiling]
+  [q message-id-ceiling maybe-send-cmd-event! cmd-event-ch]
   (fn [command-chan update-metrics]
-    (stock/reduce q
-                  (fn [chan entry]
-                    ;;This conditional guards against a new command
-                    ;;enqueued in the same directory before we've
-                    ;;read all existing files
-                    (when (< (stock/entry-id entry) message-id-ceiling)
-                      (let [{:keys [command version] :as cmdref} (entry->cmdref entry)]
-                        (async/>!! chan cmdref)
-                        (update-metrics command version)))
-                    chan)
-                  command-chan)))
+    (let [output (stock/reduce q
+                               (fn [chan entry]
+                                 ;;This conditional guards against a new command
+                                 ;;enqueued in the same directory before we've
+                                 ;;read all existing files
+                                 (when (< (stock/entry-id entry) message-id-ceiling)
+                                   (let [{:keys [command version] :as cmdref} (entry->cmdref entry)]
+                                     (async/>!! chan cmdref)
+                                     (maybe-send-cmd-event! cmdref :puppetlabs.puppetdb.command/ingested)
+                                     (update-metrics command version)))
+                                 chan)
+                               command-chan)]
+      (async/>!! cmd-event-ch {:kind ::queue-loaded})
+      output)))
 
 (defn create-or-open-stockpile
   "Opens an existing stockpile queue if one is present otherwise
   creates a new stockpile queue at `queue-dir`"
-  [queue-dir]
+  [queue-dir maybe-send-cmd-event! cmd-event-ch]
   (let [stockpile-root (kitchensink/absolute-path queue-dir)
         queue-path (get-path stockpile-root "cmd")]
     (if-let [q (and (Files/exists queue-path (make-array LinkOption 0))
                     (stock/open queue-path))]
-      [q (message-loader q (stock/next-likely-id q))]
+      [q (message-loader q (stock/next-likely-id q) maybe-send-cmd-event! cmd-event-ch)]
       (do
         (Files/createDirectories (get-path stockpile-root)
                                  (make-array FileAttribute 0))
