@@ -58,8 +58,7 @@
             [puppetlabs.kitchensink.core :as kitchensink]
             [puppetlabs.puppetdb.scf.storage-utils :as sutils]
             [clojure.set :refer :all]
-            [puppetlabs.puppetdb.time :refer [to-timestamp]]
-            [clj-time.core :refer [now interval in-millis]]
+            [puppetlabs.puppetdb.time :refer [in-millis interval now to-timestamp]]
             [puppetlabs.puppetdb.jdbc :as jdbc :refer [query-to-vec]]
             [puppetlabs.puppetdb.config :as conf]
             [puppetlabs.i18n.core :refer [trs]]
@@ -1275,9 +1274,12 @@
           "  FROM edges e"
           "  INNER JOIN certnames c ON e.certname = c.certname")
      (str "ALTER TABLE edges_transform"
-          "  ALTER COLUMN certname SET NOT NULL,"
-          "  ALTER COLUMN source SET NOT NULL,"
-          "  ALTER COLUMN target SET NOT NULL,"
+          "  ALTER COLUMN certname SET NOT NULL")
+     (str "ALTER TABLE edges_transform"
+          "  ALTER COLUMN source SET NOT NULL")
+     (str "ALTER TABLE edges_transform"
+          "  ALTER COLUMN target SET NOT NULL")
+     (str "ALTER TABLE edges_transform"
           "  ALTER COLUMN type SET NOT NULL")
      (str "DROP TABLE edges")
      (str "ALTER TABLE edges_transform RENAME TO edges")
@@ -1315,7 +1317,8 @@
 (defn rededuplicate-facts []
   (log/info (trs "[1/8] Cleaning up unreferenced facts..."))
   (jdbc/do-commands
-   "DELETE FROM facts WHERE factset_id NOT IN (SELECT id FROM factsets)")
+   "DELETE FROM facts WHERE factset_id NOT IN (SELECT id FROM factsets)"
+   "DELETE FROM facts WHERE fact_path_id NOT IN (SELECT id FROM fact_paths)")
 
   (log/info (trs "[2/8] Creating new fact storage tables..."))
   (jdbc/do-commands
@@ -1452,13 +1455,21 @@
 (defn varchar-columns-to-text []
   (jdbc/do-commands
     "alter table reports
-     alter column puppet_version type text,
+     alter column puppet_version type text"
+
+    "alter table reports
      alter column configuration_version type text"
 
     "alter table resource_events
-     alter column status type text,
-     alter column property type text,
-     alter column containing_class type text,
+     alter column status type text"
+
+    "alter table resource_events
+     alter column property type text"
+
+    "alter table resource_events
+     alter column containing_class type text"
+
+    "alter table resource_events
      alter column file type text"))
 
 (defn jsonb-facts []
